# -*- coding: utf-8 -*-

# Standard library imports

# Third-party imports

# Application-specific imports


class Identity(object):
    def __init__(self):
        pass

    def __call__(self, x, *args, **kwargs):
        return x



# -*- coding: utf-8 -*-

# Standard library imports

# Third-party imports

# Application-specific imports


class Stitch(object):
    def __init__(self):
        pass

    def __call__(self, topology, arcs, *args, **kwargs):
        self.stitched_arcs = dict()
        self.fragment_by_start = dict()
        self.fragment_by_end = dict()
        self.fragments = list()
        self.empty_index = -1
        self.topology = topology

        # Stitch empty arcs first, since they may be subsumed by other arcs.
        for i, j in zip(arcs[0::2], arcs[1::2]):
            idx = self._get_idx(i)
            arc = self.topology['arcs'][idx]
            if len(arc) < 3 and not arc[1][0] and not arc[1][1]:
                self.empty_index += 1
                arcs[self.empty_index], arcs[j] = i, arcs[self.empty_index]

        for i in arcs:
            e = self.ends(i)
            start, end = tuple(e[0]), tuple(e[1])

            if start in self.fragment_by_end:
                f = self.fragment_by_end[start]
                self.fragment_by_end.pop(f['end'], None)
                idx = sorted(list(filter(lambda x: isinstance(x, int), f.keys())))[-1] + 1
                f[idx] = i
                f['end'] = end

                # g = self.fragment_by_start.get(end, None)
                # if g is not None:
                if end in self.fragment_by_start:
                    g = self.fragment_by_start[end]
                    self.fragment_by_start.pop(g['start'], None)
                    fg = f if g == f else f + g
                    fg['start'] = f['start']
                    fg['end'] = g['end']
                    self.fragment_by_start[fg['start']] = fg
                    self.fragment_by_end[fg['end']] = fg
                else:
                    self.fragment_by_start[f['start']] = f
                    self.fragment_by_end[f['end']] = f
            else:
                if end in self.fragment_by_start:
                    f = self.fragment_by_start[end]
                    self.fragment_by_start.pop(f['start'], None)
                    self.unshift(f, i)
                    f['start'] = start
                    g = self.fragment_by_end.get(start, None)
                    if g is not None:
                        self.fragment_by_end.pop(g['end'], None)
                        gf = f if g == f else g + f
                        gf['start'] = g['start']
                        gf['end'] = f['end']
                        self.fragment_by_start[gf['start']] = gf
                        self.fragment_by_end[gf['end']] = gf
                    else:
                        self.fragment_by_start[f['start']] = f
                        self.fragment_by_end[f['end']] = f
                else:
                    f = {
                        0: i,
                        'start': tuple(start),
                        'end': tuple(end)
                    }
                    self.fragment_by_start[f['start']] = f
                    self.fragment_by_end[f['end']] = f

        self.flush(self.fragment_by_end, self.fragment_by_start)
        self.flush(self.fragment_by_start, self.fragment_by_end)

        for i in arcs:
            idx = self._get_idx(i)
            if idx not in self.stitched_arcs:
                self.fragments.append([i])

        # Convert dict to list
        fragments = list()
        for f in self.fragments:
            if 1 in f:
                fragments.append([f[0], f[1]])
            else:
                fragments.append([f[0]])

        return fragments

    @staticmethod
    def _get_idx(i):
        return ~i if i < 0 else i

    @staticmethod
    def unshift(d, value):
        sorted_keys = sorted(
            list(
                filter(lambda x: isinstance(x, int),d.keys())
            ), reverse=True)
        for k in sorted_keys:
            d[k + 1] = d[k]
        d[0] = value

    def ends(self, i):
        idx = self._get_idx(i)
        arc = self.topology['arcs'][idx]
        p_0 = arc[0]

        if 'transform' in self.topology:
            p_1 = [0, 0]
            for dp in arc:
                p_1[0] += dp[0]
                p_1[1] += dp[1]
        else:
            p_1 = arc[-1]

        return [p_1, p_0] if i < 0 else [p_0, p_1]

    def flush(self, fragment_by_end, fragment_by_start):
        for _, f in fragment_by_end.items():
            fragment_by_start.pop(f['start'], None)
            f.pop('start', None)
            f.pop('end', None)
            for _, i in f.items():
                self.stitched_arcs[self._get_idx(i)] = 1
            self.fragments.append(f)



# -*- coding: utf-8 -*-

# Standard library imports
import unittest

# Third-party imports

# Application-specific imports
from pytopojson import prequantize


class PrequantizeTestCase(unittest.TestCase):

    def setUp(self):
        self.prequantize = prequantize.Prequantize()

    def test_prequantize_returns_the_quantization_transform(self):
        quantize = self.prequantize({}, [0, 0, 1, 1], 1e4)

        self.assertDictEqual(
            quantize,
            {
                'scale': [1 / 9999, 1 / 9999],
                'translate': [0, 0]
            }
        )

    def test_prequantize_converts_coordinates_to_fixed_precision(self):
        objects = {
            'foo': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [0, 1], [0, 0]]
            }
        }

        self.prequantize(objects, [0, 0, 1, 1], 1e4)
        self.assertCountEqual(objects['foo']['arcs'], [[0, 0], [9999, 0], [0, 9999], [0, 0]])

    def test_prequantize_observes_the_quantization_parameter(self):
        objects = {
            'foo': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [0, 1], [0, 0]]
            }
        }

        self.prequantize(objects, [0, 0, 1, 1], 10)
        self.assertCountEqual(objects['foo']['arcs'], [[0, 0], [9, 0], [0, 9], [0, 0]])

    def test_prequantize_observes_the_bounding_box(self):
        objects = {
            'foo': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [0, 1], [0, 0]]
            }
        }

        self.prequantize(objects, [-1, -1, 2, 2], 10)
        self.assertCountEqual(objects['foo']['arcs'], [[3, 3], [6, 3], [3, 6], [3, 3]])

    def test_prequantize_applies_to_points_as_well_as_arcs(self):
        objects = {
            'foo': {
                'type': 'MultiPoint',
                'coordinates': [[0, 0], [1, 0], [0, 1], [0, 0]]
            }
        }

        self.prequantize(objects, [0, 0, 1, 1], 1e4)
        self.assertCountEqual(objects['foo']['coordinates'], [[0, 0], [9999, 0], [0, 9999], [0, 0]])

    def test_prequantize_skips_coincident_points_in_lines(self):
        objects = {
            'foo': {
                'type': 'LineString',
                'arcs': [[0, 0], [0.9, 0.9], [1.1, 1.1], [2, 2]]
            }
        }

        self.prequantize(objects, [0, 0, 2, 2], 3)
        self.assertCountEqual(objects['foo']['arcs'], [[0, 0], [1, 1], [2, 2]])

    def test_prequantize_skips_coincident_points_in_polygons(self):
        objects = {
            'foo': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [0.9, 0.9], [1.1, 1.1], [2, 2], [0, 0]]]
            }
        }

        self.prequantize(objects, [0, 0, 2, 2], 3)
        self.assertCountEqual(objects['foo']['arcs'], [[[0, 0], [1, 1], [2, 2], [0, 0]]])

    def test_prequantize_does_not_skip_coincident_points_in_points(self):
        objects = {
            'foo': {
                'type': 'MultiPoint',
                'coordinates': [[0, 0], [0.9, 0.9], [1.1, 1.1], [2, 2], [0, 0]]
            }
        }

        self.prequantize(objects, [0, 0, 2, 2], 3)
        self.assertCountEqual(objects['foo']['coordinates'], [[0, 0], [1, 1], [1, 1], [2, 2], [0, 0]])

    def test_prequantize_includes_closing_point_in_degenerate_lines(self):
        objects = {
            'foo': {
                'type': 'LineString',
                'arcs': [[1, 1], [1, 1], [1, 1]]
            }
        }

        self.prequantize(objects, [0, 0, 2, 2], 3)
        self.assertCountEqual(objects['foo']['arcs'], [[1, 1], [1, 1]])

    def test_prequantize_includes_closing_point_in_degenerate_polygons(self):
        objects = {
            'foo': {
                'type': 'Polygon',
                'arcs': [[[0.9, 1], [1.1, 1], [1.01, 1], [0.9, 1]]]
            }
        }

        self.prequantize(objects, [0, 0, 2, 2], 3)
        self.assertCountEqual(objects['foo']['arcs'], [[[1, 1], [1, 1], [1, 1], [1, 1]]])


# -*- coding: utf-8 -*-

# Standard library imports
import unittest

# Third-party imports

# Application-specific imports
from pytopojson import geometry


class GeometryTestCase(unittest.TestCase):
    def setUp(self):
        self.geometry = geometry.Geometry()

    def test_geometry_replaces_LineString_Feature_with_LineString_Geometry(self):
        geom = self.geometry({
            'foo': {
                'type': 'Feature',
                'geometry': {
                    'type': 'LineString',
                    'coordinates': [[0, 0]]
                }
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'LineString',
                    'arcs': [[0, 0]]
                }
            }
        )

    def test_geometry_replaces_GeometryCollection_Feature_with_GeometryCollection(self):
        geom = self.geometry({
            'foo': {
                'type': 'Feature',
                'geometry': {
                    'type': 'GeometryCollection',
                    'geometries': [{
                        'type': 'LineString',
                        'coordinates': [[0, 0]]
                    }]
                }
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'GeometryCollection',
                    'geometries': [{
                        'type': 'LineString',
                        'arcs': [[0, 0]]
                    }]
                }
            }
        )

    def test_geometry_replaces_FeatureCollection_with_GeometryCollection(self):
        geom = self.geometry({
            'foo': {
                'type': 'FeatureCollection',
                'features': [{
                    'type': 'Feature',
                    'geometry': {
                        'type': 'LineString',
                        'coordinates': [[0, 0]]
                    }
                }]
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'GeometryCollection',
                    'geometries': [{
                        'type': 'LineString',
                        'arcs': [[0, 0]]
                    }]
                }
            }
        )

    def test_geometry_replaces_Feature_with_null_Geometry_with_null_type_Geometry(self):
        geom = self.geometry({
            'foo': {
                'type': 'Feature',
                'geometry': None
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': None
                }
            }
        )

    def test_geometry_replaces_top_level_null_Geometry_with_null_type_Geometry(self):
        geom = self.geometry({
            'foo': None
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': None
                }
            }
        )

    def test_geometry_replaces_null_Geometry_in_GeometryCollection_with_null_type_Geometry(self):
        geom = self.geometry({
            'foo': {
                'type': 'GeometryCollection',
                'geometries': [None]
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'GeometryCollection',
                    'geometries': [{
                        'type': None
                    }]
                }
            }
        )

    def test_geometry_preserves_id(self):
        geom = self.geometry({
            'foo': {
                'id': 'foo',
                'type': 'Feature',
                'geometry': {
                    'type': 'LineString',
                    'coordinates': [[0, 0]]
                }
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'id': 'foo',
                    'type': 'LineString',
                    'arcs': [[0, 0]]
                }
            }
        )

    def test_geometry_preserves_properties_if_non_empty(self):
        geom = self.geometry({
            'foo': {
                'properties': {
                    'foo': 42
                },
                'type': 'Feature',
                'geometry': {
                    'type': 'LineString',
                    'coordinates': [[0, 0]]
                }
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'properties': {
                        'foo': 42
                    },
                    'type': 'LineString',
                    'arcs': [[0, 0]]
                }
            }
        )

    def test_geometry_applies_a_shallow_copy_for_properties(self):
        input = {
            'foo': {
                'properties': {
                    'foo': 42
                },
                'type': 'Feature',
                'geometry': {
                    'type': 'LineString',
                    'coordinates': [[0, 0]]
                }
            }
        }

        geom = self.geometry(input)

        self.assertDictEqual(input['foo']['properties'], geom['foo']['properties'])

    def test_geometry_deletes_empty_properties(self):
        geom = self.geometry({
            'foo': {
                'properties': {},
                'type': 'Feature',
                'geometry': {
                    'type': 'LineString',
                    'coordinates': [[0, 0]]
                }
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'LineString',
                    'arcs': [[0, 0]]
                }
            }
        )

    def test_geometry_does_not_convert_singular_multipoints_to_points(self):
        geom = self.geometry({
            'foo': {
                'type': 'MultiPoint',
                'coordinates': [[0, 0]]
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'MultiPoint',
                    'coordinates': [[0, 0]]
                }
            }
        )

    def test_geometry_does_not_convert_empty_multipoints_to_null(self):
        geom = self.geometry({
            'foo': {
                'type': 'MultiPoint',
                'coordinates': []
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'MultiPoint',
                    'coordinates': []
                }
            }
        )

    def test_geometry_does_not_convert_singular_multilines_to_lines(self):
        geom = self.geometry({
            'foo': {
                'type': 'MultiLineString',
                'coordinates': [[[0, 0], [0, 1]]]
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'MultiLineString',
                    'arcs': [[[0, 0], [0, 1]]]
                }
            }
        )

    def test_geometry_does_not_convert_empty_lines_to_null(self):
        geom = self.geometry({
            'foo': {
                'type': 'LineString',
                'coordinates': []
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'LineString',
                    'arcs': []
                }
            }
        )

    def test_geometry_does_not_convert_empty_multilines_to_null(self):
        geom = self.geometry({
            'foo': {
                'type': 'MultiLineString',
                'coordinates': []
            },
            'bar': {
                'type': 'MultiLineString',
                'coordinates': [[]]
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'MultiLineString',
                    'arcs': []
                },
                'bar': {
                    'type': 'MultiLineString',
                    'arcs': [[]]
                }
            }
        )

    def test_geometry_does_not_strip_empty_rings_in_polygons(self):
        geom = self.geometry({
            'foo': {
                'type': 'Polygon',
                'coordinates': [[[0, 0], [1, 0], [1, 1], [0, 0]], []]
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'Polygon',
                    'arcs': [[[0, 0], [1, 0], [1, 1], [0, 0]], []]
                }
            }
        )

    def test_geometry_does_not_strip_empty_lines_in_multilines(self):
        geom = self.geometry({
            'foo': {
                'type': 'MultiLineString',
                'coordinates': [[[0, 0], [1, 0], [1, 1], [0, 0]], [], [[0, 0], [1, 0]]]
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'MultiLineString',
                    'arcs': [[[0, 0], [1, 0], [1, 1], [0, 0]], [], [[0, 0], [1, 0]]]
                }
            }
        )

    def test_geometry_does_not_convert_empty_polygons_to_null(self):
        geom = self.geometry({
            'foo': {
                'type': 'Polygon',
                'coordinates': []
            },
            'bar': {
                'type': 'Polygon',
                'coordinates': [[]]
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'Polygon',
                    'arcs': []
                },
                'bar': {
                    'type': 'Polygon',
                    'arcs': [[]]
                }
            }
        )

    def test_geometry_does_not_strip_empty_polygons_in_multipolygons(self):
        geom = self.geometry({
            'foo': {
                'type': 'MultiPolygon',
                'coordinates': [[[[0, 0], [1, 0], [1, 1], [0, 0]], []], [], [[]]]
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'MultiPolygon',
                    'arcs': [[[[0, 0], [1, 0], [1, 1], [0, 0]], []], [], [[]]]
                }
            }
        )

    def test_geometry_does_not_convert_singular_multipolygons_to_polygons(self):
        geom = self.geometry({
            'foo': {
                'type': 'MultiPolygon',
                'coordinates': [[[[0, 0], [0, 1], [1, 0], [0, 0]]]]
            }
        })

        self.assertDictEqual(
            geom,
            {
                'foo': {
                    'type': 'MultiPolygon',
                    'arcs': [[[[0, 0], [0, 1], [1, 0], [0, 0]]]]
                }
            }
        )


# -*- coding: utf-8 -*-

# Standard library imports
import argparse
import json
import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(__file__)))

# Third-party imports

# Application-specific imports
from pytopojson import topology


__version__ = '0.0.1'


def read(specifiers):
    objs = dict()

    for specifier in specifiers:
        i = specifier.find('=')
        if 0 <= i:
            name, file = specifier.split('=')
        else:
            file = specifier
            filename = os.path.basename(file)
            name, _ = os.path.splitext(filename)

        if name in objs:
            print(f'\n  error: object {name} is not unique\n')

        else:
            objs[name] = read_object(file)

    return objs


def read_object(filepath):
    with open(filepath, 'r') as f:
        data = f.read()
    return json.loads(data)


def read_newline_delimited_object():
    pass


def write(data, output):
    if isinstance(output, str):
        with open(output, 'w') as dst:
            json.dump(data, dst)
    else:
        json.dump(data, output)


if __name__ == '__main__':
    # Create OptionParser object and set options
    parser = argparse.ArgumentParser(description='Converts GeoJSON features to TopoJSON objects.')

    parser.add_argument('-o', '--out', dest='file',
                        help='output file name; defaults to “-” for stdout',
                        default='-')
    # parser.add_argument('-n', '--newline-delimited',
    #                     help='accept newline-delimited JSON')
    parser.add_argument('-q', '--quantization', dest='count', type=float,
                        help='pre-quantization parameter; 0 disables quantization',
                        default=0)
    parser.add_argument('geojsons', metavar='[name=]file', type=str, nargs='+',
                        help='file with GeoJSON data')
    parser.add_argument('-v', '--version', action='version',
                        version=f'%(prog)s v{__version__}')

    # Parse and store the command-line arguments in dictionary
    opts = parser.parse_args()
    kwargs = vars(opts)

    # Output destination
    out = kwargs['file'] if kwargs['file'] != '-' else sys.stdout

    # Quantization
    quant = {'quantization': kwargs['count']}
    quant = quant if quant['quantization'] != 0 else dict()

    # Read files and compute topology
    objects = read(kwargs['geojsons'])
    topology = topology.Topology()
    topo = topology(objects, **quant)

    # Write TopoJSON
    write(topo, out)



# -*- coding: utf-8 -*-

# Standard library imports
import unittest

# Third-party imports

# Application-specific imports
from pytopojson import extract


class ExtractTestCase(unittest.TestCase):

    def setUp(self):
        self.extract = extract.Extract()

    def test_extract_copies_coordinates_sequentially_into_a_buffer(self):
        topology = self.extract({
            'foo': {
                'type': "LineString",
                'arcs': [[0, 0], [1, 0], [2, 0]]
            },
            'bar': {
                'type': "LineString",
                'arcs': [[0, 0], [1, 0], [2, 0]]
            }
        })

        self.assertCountEqual(topology['coordinates'], [[0, 0], [1, 0], [2, 0], [0, 0], [1, 0], [2, 0]])

    def test_extract_does_not_copy_point_geometries_into_the_coordinate_buffer(self):
        topology = self.extract({
            'foo': {
                'type': 'Point',
                'arcs': [0, 0]
            },
            'bar': {
                'type': 'MultiPoint',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            }
        })

        self.assertCountEqual(topology['coordinates'], [])
        self.assertCountEqual(topology['objects']['foo']['arcs'], [0, 0])
        self.assertCountEqual(topology['objects']['bar']['arcs'], [[0, 0], [1, 0], [2, 0]])

    def test_extract_includes_closing_coordinates_in_polygons(self):
        topology = self.extract({
            'foo': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [2, 0], [0, 0]]]
            }
        })

        self.assertCountEqual(topology['coordinates'], [[0, 0], [1, 0], [2, 0], [0, 0]])

    def test_extract_represents_lines_as_contiguous_slices_of_the_coordinate_buffer(self):
        topology = self.extract({
            'foo': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            },
            'bar': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            }
        })

        self.assertDictEqual(
            topology['objects'],
            {
                'foo': {
                    'type': 'LineString',
                    'arcs': [0, 2]
                },
                'bar': {
                    'type': 'LineString',
                    'arcs': [3, 5]
                }
            }
        )

    def test_extract_represents_rings_as_contiguous_slices_of_the_coordinate_buffer(self):
        topology = self.extract({
            'foo': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [2, 0], [0, 0]]]
            },
            'bar': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [2, 0], [0, 0]]]
            }
        })

        self.assertDictEqual(
            topology['objects'],
            {
                'foo': {
                    'type': 'Polygon',
                    'arcs': [[0, 3]]
                },
                'bar': {
                    'type': 'Polygon',
                    'arcs': [[4, 7]]
                }
            }
        )

    def test_extract_exposes_the_constructed_lines_and_rings_in_the_order_of_construction(self):
        topology = self.extract({
            'line': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            },
            'multiline': {
                'type': 'MultiLineString',
                'arcs': [[[0, 0], [1, 0], [2, 0]]]
            },
            'polygon': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [2, 0], [0, 0]]]
            }
        })

        self.assertCountEqual(topology['lines'], [[0, 2], [3, 5]])
        self.assertCountEqual(topology['rings'], [[6, 9]])

    def test_extract_supports_nested_geometry_collections(self):
        topology = self.extract({
            'foo': {
                'type': 'GeometryCollection',
                'geometries': [{
                    'type': 'GeometryCollection',
                    'geometries': [{
                        'type': 'LineString',
                        'arcs': [[0, 0], [0, 1]]
                    }]
                }]
            }
        })

        self.assertDictEqual(
            topology['objects']['foo'],
            {
                'type': 'GeometryCollection',
                'geometries': [{
                    'type': 'GeometryCollection',
                    'geometries': [{
                        'type': 'LineString',
                        'arcs': [0, 1]
                    }]
                }]
            }

        )


# -*- coding: utf-8 -*-

# Standard library imports
import argparse
import json
import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(__file__)))

# Third-party imports

# Application-specific imports
from pytopojson import feature


__version__ = '0.0.1'


def read(input_):
    if isinstance(input_, str):
        with open(input_, 'r') as src:
            data = json.load(src)
    else:
        data = json.load(input_)

    return data


def write_list(topology_):
    for name in topology_['objects']:
        print(name)


def read_newline_delimited_object():
    pass


def write(topology_, objects):
    feat = feature.Feature()
    for specifier in objects:
        i = specifier.find('=')
        if 0 <= i:
            name, file_ = specifier.split('=')
        else:
            file_ = specifier
            filename = os.path.basename(file_)
            name, _ = os.path.splitext(filename)

        if name not in topology_['objects']:
            print(f'\n  error: object {name} not found\n')
            return

        write_feature(file_, feat(topology, name))


def write_feature(output, feat):
    if output == '-':
        json.dump(feat, sys.stdout)
        sys.stdout.write('\n')
    else:
        with open(output, 'w') as dst:
            json.dump(feat, dst)
            dst.write('\n')


if __name__ == '__main__':
    # Create OptionParser object and set options
    parser = argparse.ArgumentParser(description='Converts TopoJSON objects to GeoJSON features.')

    parser.add_argument('-i', '--in', dest='file',
                        help='input topology file name; defaults to "-" for stdin',
                        default='-')
    parser.add_argument('-l', '--list', action='store_true', default=False,
                        help='list the object names on the input topology')
    parser.add_argument('objects', metavar='[geometry_type=]file', type=str, nargs='+',
                        help='output with GeoJSON data for given geometry in input')
    # parser.add_argument('-n', '--newline-delimited',
    #                     help='output newline-delimited JSON')
    parser.add_argument('-v', '--version', action='version',
                        version=f'%(prog)s v{__version__}')

    # Parse and store the command-line arguments in dictionary
    opts = parser.parse_args()
    kwargs = vars(opts)

    # Read input file
    input_ = kwargs['file'] if kwargs['file'] != '-' else sys.stdin
    topology = read(input_)

    if kwargs['list']:
        write_list(topology)
    else:
        write(topology, kwargs['objects'])



# -*- coding: utf-8 -*-

# Standard library imports
import unittest

# Third-party imports

# Application-specific imports
from pytopojson import delta


class DeltaTestCase(unittest.TestCase):

    def setUp(self):
        self.delta = delta.Delta()

    def test_delta_converts_arcs_to_delta_encoding(self):
        d = self.delta([
            [[0, 0], [9999, 0], [0, 9999], [0, 0]]
        ])

        self.assertCountEqual(
            d,
            [
                [[0, 0], [9999, 0], [-9999, 9999], [0, -9999]]
            ]
        )

    def test_delta_skips_coincident_points(self):
        d = self.delta([
            [[0, 0], [9999, 0], [9999, 0], [0, 9999], [0, 0]]
        ])

        self.assertCountEqual(
            d,
            [
                [[0, 0], [9999, 0], [-9999, 9999], [0, -9999]]
            ]
        )

    def test_delta_preserves_at_least_two_positions(self):
        d = self.delta([
            [[12345, 12345], [12345, 12345], [12345, 12345], [12345, 12345]]
        ])

        self.assertCountEqual(
            d,
            [
                [[12345, 12345], [0, 0]]
            ]
        )


# -*- coding: utf-8 -*-

# Standard library imports
import argparse
import json
import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(__file__)))

# Third-party imports

# Application-specific imports
from pytopojson import quantize


__version__ = '0.0.1'


def read(input_):
    if isinstance(input_, str):
        with open(input_, 'r') as src:
            return json.load(src)
    else:
        return json.load(sys.stdin)


def quantize_topology(topology, quantization):
    quantize_ = quantize.Quantize()
    return quantize_(topology, quantization)


def write(output, topology):
    if isinstance(output, str):
        with open(output, 'w') as dst:
            json.dump(topology, dst)
            dst.write('\n')
    else:
        json.dump(topology, sys.stdout)
        sys.stdout.write('\n')


def _valid_quantization_parameter(arg):
    try:
        f = float(arg)
    except ValueError:
        raise argparse.ArgumentTypeError("Must be a floating point number")
    if f <= 2.0:
        raise argparse.ArgumentTypeError(f'Invalid quantization parameter {arg}')
    return f


if __name__ == '__main__':
    # Create OptionParser object and set options
    parser = argparse.ArgumentParser(description='Quantizes TopoJSON.')

    parser.add_argument('-i', '--in', dest='input', default='-',
                        help='input topology file name; defaults to "-" for stdin')
    parser.add_argument('-o', '--out', dest='output', default='-',
                        help='output topology file name; defaults to "-" for stdout')
    parser.add_argument('-q', '--quantization', dest='quantization',
                        type=_valid_quantization_parameter, required=True,
                        help='quantization value')
    parser.add_argument('-v', '--version', action='version',
                        version=f'%(prog)s v{__version__}')

    # Parse and store the command-line arguments in dictionary
    opts = parser.parse_args()
    kwargs = vars(opts)

    # Read input file
    input_ = kwargs['input'] if kwargs['input'] != '-' else sys.stdin
    output = kwargs['output'] if kwargs['output'] != '-' else sys.stdout

    topology = read(input_)
    quantized = quantize_topology(topology, kwargs['quantization'])
    write(output, quantized)



# -*- coding: utf-8 -*-

# Standard library imports
import json
import subprocess
import unittest

# Third-party imports

# Application-specific imports


class TopoQuantizeTestCase(unittest.TestCase):
    def topoquantize(self, command, result):
        command = ' '.join(command.split()).split(' ')
        p = subprocess.run(['python', 'bin/topoquantize.py', *command],
                           capture_output=True)
        out, err = p.stdout, p.stderr
        actual = json.loads(out)
        with open(result, 'r') as f:
            expected = json.load(f)

        self.assertDictEqual(actual, expected)

    def test_polygon_quantization(self):
        # self.topoquantize('-q 1e4 < tests/client/topojson/polygon.json',
        self.topoquantize('-q 1e4 -i tests/client/topojson/polygon.json',
                          'tests/client/topojson/polygon-q1e4.json')

    def test_polygon_quantization_2(self):
        # self.topoquantize('-q 1e5 < tests/client/topojson/polygon.json',
        self.topoquantize('-q 1e5 -i tests/client/topojson/polygon.json',
                          'tests/client/topojson/polygon-q1e5.json')

    def test_polygon_mercator_quantization(self):
        # self.topoquantize('-q 1e4 < tests/client/topojson/polygon-mercator.json',
        self.topoquantize('-q 1e4 -i tests/client/topojson/polygon-mercator.json',
                          'tests/client/topojson/polygon-mercator-q1e4.json')

    def test_polygon_mercator_quantization_2(self):
        # self.topoquantize('-q 1e5 < tests/client/topojson/polygon-mercator.json',
        self.topoquantize('-q 1e5 -i tests/client/topojson/polygon-mercator.json',
                          'tests/client/topojson/polygon-mercator-q1e5.json')

    def test_point_quantization(self):
        # self.topoquantize('-q 1e5 < tests/client/topojson/point.json',
        self.topoquantize('-q 1e5 -i tests/client/topojson/point.json',
                          'tests/client/topojson/point-q1e5.json')

    def test_points_quantization(self):
        # self.topoquantize('-q 1e5 < tests/client/topojson/points.json',
        self.topoquantize('-q 1e5 -i tests/client/topojson/points.json',
                          'tests/client/topojson/points-q1e5.json')



# -*- coding: utf-8 -*-

# Standard library imports

# Third-party imports

# Application-specific imports
from pytopojson import (
    transform,
)


class Reverse(object):
    def __init__(self):
        pass

    def __call__(self, array, n, *args, **kwargs):
        j = len(array)
        i = j - n

        j -= 1
        while i < j:
            array[i], array[j] = array[j], array[i]
            i += 1
            j -= 1


class Object(object):
    def __init__(self):
        self.reverse = Reverse()
        self.transform = transform.Transform()
        self.transform_point = None
        self.arcs = None

    def __call__(self, topology, o, *args, **kwargs):
        self.transform_point = self.transform(topology.get('transform', None))
        self.arcs = topology['arcs']
        return self.geometry(o)

    def arc(self, i, points):
        if any(points):
            points.pop()

        idx = ~i if i < 0 else i
        arcs = self.arcs[idx]
        for k, arc in enumerate(arcs):
            points.append(self.transform_point(arc,  k))

        if i < 0:
            self.reverse(points, len(arcs))

    def point(self, p):
        return self.transform_point(p)

    def line(self, arcs):
        points = list()
        for arc in arcs:
            self.arc(arc, points)

        # This should never happen per the specification.
        if len(points) < 2:
            points.append(points[0])

        return points

    def ring(self, arcs):
        points = self.line(arcs)
        # This may happen if an arc has only two points.
        while len(points) < 4:
            points.append(points[0])
        return points

    def polygon(self, arcs):
        return list(map(lambda x: self.ring(x), arcs))

    def geometry(self, o):
        _type = o.get('type', None)

        if _type == 'GeometryCollection':
            return {
                'type': _type,
                'geometries': list(map(lambda x: self.geometry(x), o['geometries']))
            }
        elif _type == 'Point':
            coordinates = self.point(o['coordinates'])
        elif _type == 'MultiPoint':
            coordinates = list(map(lambda x: self.point(x), o['coordinates']))
        elif _type == 'LineString':
            coordinates = self.line(o['arcs'])
        elif _type == 'MultiLineString':
            coordinates = list(map(lambda x: self.line(x), o['arcs']))
        elif _type == 'Polygon':
            coordinates = self.polygon(o['arcs'])
        elif _type == 'MultiPolygon':
            coordinates = list(map(lambda x: self.polygon(x), o['arcs']))
        else:
            return None

        return {
            'type': _type,
            'coordinates': coordinates
        }


class Feature(object):
    def __init__(self):
        self.object = Object()
        self.reverse = Reverse()
        self.transform = transform.Transform()

    def __call__(self, topology, o, *args, **kwargs):
        if isinstance(o, str):
            o = topology['objects'][o]
        if o.get('type', None) == 'GeometryCollection':
            return {
                'type': 'FeatureCollection',
                'features': list(map(lambda x: self.feature(topology, x),
                                     o['geometries']))
            }
        else:
            return self.feature(topology, o)

    def feature(self, topology, o):
        feat = {
            'id': o.get('id', None),
            'bbox': o.get('bbox', None),
            'type': 'Feature',
            'properties': o.get('properties', dict()),
            'geometry': self.object(topology, o)
        }

        for k in ('id', 'bbox'):
            if feat[k] is None:
                del feat[k]

        return feat


# -*- coding: utf-8 -*-

# Standard library imports
import json
import subprocess
import unittest

# Third-party imports

# Application-specific imports
from tests import in_delta


class Geo2TopoTestCase(unittest.TestCase):
    def geo2topo(self, output, options=''):
        options = ' '.join(options.split()).split(' ')
        p = subprocess.run(['python', 'bin/geo2topo.py', *options],
                           capture_output=True)
        out, err = p.stdout, p.stderr
        actual = json.loads(out.decode('utf8').strip().replace("'", '"'))
        with open(output, 'r') as f:
            expected = json.load(f)

        self.assertTrue(in_delta.in_delta(actual, expected))

    def test_polygons(self):
        self.geo2topo('tests/server/topojson/polygon-no-quantization.json',
                      'polygon=tests/server/geojson/polygon-clockwise.json')

    def test_quantized_polygons(self):
        self.geo2topo('tests/server/topojson/polygon.json',
                      '-q 1E4 polygon=tests/server/geojson/polygon-clockwise.json')

    def test_empty_geometries(self):
        self.geo2topo('tests/server/topojson/empty.json',
                      'multilinestring=tests/server/geojson/empty-multilinestring.json'
                      ' multipoint=tests/server/geojson/empty-multipoint.json'
                      ' multipolygon=tests/server/geojson/empty-multipolygon.json'
                      ' multipolygon2=tests/server/geojson/empty-multipolygon2.json'
                      ' polygon=tests/server/geojson/empty-polygon.json')



# -*- coding: utf-8 -*-

# Standard library imports
import unittest

# Third-party imports

# Application-specific imports
from pytopojson import cut, extract, dedup


class CutTestCase(unittest.TestCase):

    def setUp(self):
        self.cut = cut.Cut()
        self.dedup = dedup.Dedup()
        self.extract = extract.Extract()

    def test_dedup_exact_duplicate_lines_abc_and_abc_share_an_arc(self):
        topology = self.dedup(self.cut(self.extract({
            'abc': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            },
            'abc2': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            }
        })))

        self.assertDictEqual({
            'abc': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 2}
            },
            'abc2': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 2}
            }
        }, topology['objects'])

    def test_dedup_reversed_duplicate_lines_abc_and_cba_share_an_arc(self):
        topology = self.dedup(self.cut(self.extract({
            'abc': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            },
            'cba': {
                'type': 'LineString',
                'arcs': [[2, 0], [1, 0], [0, 0]]
            }
        })))

        self.assertDictEqual({
            'abc': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 2}
            },
            'cba': {
                'type': 'LineString',
                'arcs': {0: 2, 1: 0}
            }
        }, topology['objects'])

    def test_dedup_exact_duplicate_rings_abca_and_abca_share_an_arc(self):
        topology = self.dedup(self.cut(self.extract({
            'abca': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [2, 0], [0, 0]]]
            },
            'abca2': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [2, 0], [0, 0]]]
            }
        })))

        self.assertDictEqual({
            'abca': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            },
            'abca2': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            }
        }, topology['objects'])

    def test_dedup_reversed_duplicate_rings_acba_and_abca_share_an_arc(self):
        topology = self.dedup(self.cut(self.extract({
            'abca': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [2, 0], [0, 0]]]
            },
            'acba': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [2, 0], [1, 0], [0, 0]]]
            }
        })))

        self.assertDictEqual({
            'abca': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            },
            'acba': {
                'type': 'Polygon',
                'arcs': [{0: 3, 1: 0}]
            }
        }, topology['objects'])

    def test_dedup_rotated_duplicate_rings_bcab_and_abca_share_an_arc(self):
        topology = self.dedup(self.cut(self.extract({
            'abca': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [2, 0], [0, 0]]]
            },
            'bcab': {
                'type': 'Polygon',
                'arcs': [[[1, 0], [2, 0], [0, 0], [1, 0]]]
            }
        })))

        self.assertDictEqual({
            'abca': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            },
            'bcab': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            }
        }, topology['objects'])

    def test_dedup_ring_abca_and_line_abca_have_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'abcaLine': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0], [0, 0]]
            },
            'abcaPolygon': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [2, 0], [0, 0]]]
            }
        })))

        self.assertDictEqual({
            'abcaLine': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 3}
            },
            'abcaPolygon': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            }
        }, topology['objects'])

    def test_dedup_ring_bcab_and_line_abca_have_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'abcaLine': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0], [0, 0]]
            },
            'bcabPolygon': {
                'type': 'Polygon',
                'arcs': [[[1, 0], [2, 0], [0, 0], [1, 0]]]
            }
        })))

        self.assertDictEqual({
            'abcaLine': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 3}
            },
            'bcabPolygon': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            }
        }, topology['objects'])

    def test_dedup_ring_abca_and_line_bcab_have_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'bcabLine': {
                'type': 'LineString',
                'arcs': [[1, 0], [2, 0], [0, 0], [1, 0]]
            },
            'abcaPolygon': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [2, 0], [0, 0]]] # rotated to BCAB
            }
        })))

        self.assertDictEqual({
            'bcabLine': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 3}
            },
            'abcaPolygon': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            }
        }, topology['objects'])

    def test_dedup_when_an_old_arc_abc_extends_a_new_arc_ab_abc_is_cut_into_ab_bc(self):
        topology = self.dedup(self.cut(self.extract({
            'abc': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            },
            'ab': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0]]
            }
        })))

        self.assertDictEqual({
            'abc': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 1, 1: 2}}
            },
            'ab': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1}
            }
        }, topology['objects'])

    def test_dedup_when_a_reversed_old_arc_cba_extends_a_new_arc_ab_cba_is_cut_into_cb_ba(self):
        topology = self.dedup(self.cut(self.extract({
            'cba': {
                'type': 'LineString',
                'arcs': [[2, 0], [1, 0], [0, 0]]
            },
            'ab': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0]]
            }
        })))

        self.assertDictEqual({
            'cba': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 1, 1: 2}}
            },
            'ab': {
                'type': 'LineString',
                'arcs': {0: 2, 1: 1}
            }
        }, topology['objects'])

    def test_dedup_when_a_new_arc_ade_shares_its_start_with_an_old_arc_abc_there_are_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'ade': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            },
            'abc': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 1], [2, 1]]
            }
        })))

        self.assertDictEqual({
            'ade': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 2}
            },
            'abc': {
                'type': 'LineString',
                'arcs': {0: 3, 1: 5}
            }
        }, topology['objects'])

    def test_dedup_ring_aba_has_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'aba': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [0, 0]]]
            }
        })))

        self.assertDictEqual({
            'aba': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 2}]
            }
        }, topology['objects'])

    def test_dedup_ring_aa_has_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'aa': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [0, 0]]]
            }
        })))

        self.assertDictEqual({
            'aa': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 1}]
            }
        }, topology['objects'])

    def test_dedup_degenerate_ring_a_has_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'a': {
                'type': 'Polygon',
                'arcs': [[[0, 0]]]
            }
        })))

        self.assertDictEqual({
            'a': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 0}]
            }
        }, topology['objects'])

    def test_dedup_when_a_new_line_dec_shares_its_end_with_an_old_line_abc_there_are_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'abc': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            },
            'dec': {
                'type': 'LineString',
                'arcs': [[0, 1], [1, 1], [2, 0]]
            }
        })))

        self.assertDictEqual({
            'abc': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 2}
            },
            'dec': {
                'type': 'LineString',
                'arcs': {0: 3, 1: 5}
            }
        }, topology['objects'])

    def test_dedup_when_a_new_line_abc_extends_an_old_line_ab_abc_is_cut_into_ab_bc(self):
        topology = self.dedup(self.cut(self.extract({
            'ab': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0]]
            },
            'abc': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            }
        })))

        self.assertDictEqual({
            'ab': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1}
            },
            'abc': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 3, 1: 4}}
            }
        }, topology['objects'])

    def test_dedup_when_a_new_line_abc_extends_a_reversed_old_line_ba_abc_is_cut_into_ab_bc(self):
        topology = self.dedup(self.cut(self.extract({
            'ba': {
                'type': 'LineString',
                'arcs': [[1, 0], [0, 0]]
            },
            'abc': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            }
        })))

        self.assertDictEqual({
            'ba': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1}
            },
            'abc': {
                'type': 'LineString',
                'arcs': {0: 1, 1: 0, 'next': {0: 3, 1: 4}}
            }
        }, topology['objects'])

    def test_dedup_when_a_new_line_starts_bc_in_the_middle_of_an_old_line_abc_abc_is_cut_into_ab_bc(self):
        topology = self.dedup(self.cut(self.extract({
            'abc': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            },
            'bc': {
                'type': 'LineString',
                'arcs': [[1, 0], [2, 0]]
            }
        })))

        self.assertDictEqual({
            'abc': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 1, 1: 2}}
            },
            'bc': {
                'type': 'LineString',
                'arcs': {0: 1, 1: 2}
            }
        }, topology['objects'])

    def test_dedup_when_a_new_line_bc_starts_in_the_middle_of_a_reversed_old_line_cba_cba_is_cut_into_cb_ba(
            self):
        topology = self.dedup(self.cut(self.extract({
            'cba': {
                'type': 'LineString',
                'arcs': [[2, 0], [1, 0], [0, 0]]
            },
            'bc': {
                'type': 'LineString',
                'arcs': [[1, 0], [2, 0]]
            }
        })))

        self.assertDictEqual({
            'cba': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 1, 1: 2}}
            },
            'bc': {
                'type': 'LineString',
                'arcs': {0: 1, 1: 0}
            }
        }, topology['objects'])

    def test_dedup_when_a_new_line_abd_deviates_from_an_old_line_abc_abd_is_cut_into_ab_bd_and_abc_is_cut_into_ab_bc(
            self):
        topology = self.dedup(self.cut(self.extract({
            'abc': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            },
            'abd': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [3, 0]]
            }
        })))

        self.assertDictEqual({
            'abc': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 1, 1: 2}}
            },
            'abd': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 4, 1: 5}}
            }
        }, topology['objects'])

    def test_dedup_when_a_new_line_abd_deviates_from_a_reversed_old_line_cba_cba_is_cut_into_cb_ba_and_abd_is_cut_into_ab_bd(
            self):
        topology = self.dedup(self.cut(self.extract({
            'cba': {
                'type': 'LineString',
                'arcs': [[2, 0], [1, 0], [0, 0]]
            },
            'abd': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [3, 0]]
            }
        })))

        self.assertDictEqual({
            'cba': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 1, 1: 2}}
            },
            'abd': {
                'type': 'LineString',
                'arcs': {0: 2, 1: 1, 'next': {0: 4, 1: 5}}
            }
        }, topology['objects'])

    def test_dedup_when_a_new_line_dbc_merges_into_an_old_line_abc_dbc_is_cut_into_db_bc_and_abc_is_cut_into_ab_bc(
            self):
        topology = self.dedup(self.cut(self.extract({
            'abc': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            },
            'dbc': {
                'type': 'LineString',
                'arcs': [[3, 0], [1, 0], [2, 0]]
            }
        })))

        self.assertDictEqual({
            'abc': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 1, 1: 2}}
            },
            'dbc': {
                'type': 'LineString',
                'arcs': {0: 3, 1: 4, 'next': {0: 1, 1: 2}}
            }
        }, topology['objects'])

    def test_dedup_when_a_new_line_dbc_merges_into_a_reversed_old_line_cba_dbc_is_cut_into_db_bc_and_cba_is_cut_into_cb_ba(
            self):
        topology = self.dedup(self.cut(self.extract({
            'cba': {
                'type': 'LineString',
                'arcs': [[2, 0], [1, 0], [0, 0]]
            },
            'dbc': {
                'type': 'LineString',
                'arcs': [[3, 0], [1, 0], [2, 0]]
            }
        })))

        self.assertDictEqual({
            'cba': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 1, 1: 2}}
            },
            'dbc': {
                'type': 'LineString',
                'arcs': {0: 3, 1: 4, 'next': {0: 1, 1: 0}}
            }
        }, topology['objects'])

    def test_dedup_when_a_new_line_dbe_shares_a_single_midpoint_with_an_old_line_abc_dbe_is_cut_into_db_be_and_abc_is_cut_into_ab_bc(
            self):
        topology = self.dedup(self.cut(self.extract({
            'abc': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0]]
            },
            'dbc': {
                'type': 'LineString',
                'arcs': [[0, 1], [1, 0], [2, 1]]
            }
        })))

        self.assertDictEqual({
            'abc': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 1, 1: 2}}
            },
            'dbc': {
                'type': 'LineString',
                'arcs': {0: 3, 1: 4, 'next': {0: 4, 1: 5}}
            }
        }, topology['objects'])

    def test_dedup_when_a_new_line_abde_skips_a_point_with_an_old_line_abcde_abde_is_cut_into_ab_bd_de_and_abcde_is_cut_into_ab_bcd_de(
            self):
        topology = self.dedup(self.cut(self.extract({
            'abcde': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0], [3, 0], [4, 0]]
            },
            'abde': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [3, 0], [4, 0]]
            }
        })))

        self.assertDictEqual({
            'abcde': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 1, 1: 3, 'next': {0: 3, 1: 4}}}
            },
            'abde': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 6, 1: 7, 'next': {0: 3, 1: 4}}}
            }
        }, topology['objects'])

    def test_dedup_when_a_new_line_abde_skips_a_point_with_a_reversed_old_line_edcba_abde_is_cut_into_ab_bd_de_and_edcba_is_cut_into_ed_dcb_ba(
            self):
        topology = self.dedup(self.cut(self.extract({
            'edcba': {
                'type': 'LineString',
                'arcs': [[4, 0], [3, 0], [2, 0], [1, 0], [0, 0]]
            },
            'abde': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [3, 0], [4, 0]]
            }
        })))

        self.assertDictEqual({
            'edcba': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 1, 1: 3, 'next': {0: 3, 1: 4}}}
            },
            'abde': {
                'type': 'LineString',
                'arcs': {0: 4, 1: 3, 'next': {0: 6, 1: 7, 'next': {0: 1, 1: 0}}}
            }
        }, topology['objects'])

    def test_dedup_when_a_line_abcdbe_self_intersects_with_its_middle_it_is_not_cut(self):
        topology = self.dedup(self.cut(self.extract({
            'abcdbe': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0], [3, 0], [1, 0], [4, 0]]
            }
        })))

        self.assertDictEqual({
            'abcdbe': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 5}
            }
        }, topology['objects'])

    def test_dedup_when_a_line_abacd_self_intersects_with_its_start_it_is_cut_into_aba_acd(self):
        topology = self.dedup(self.cut(self.extract({
            'abacd': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [0, 0], [3, 0], [4, 0]]
            }
        })))

        self.assertDictEqual({
            'abacd': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 2, 'next': {0: 2, 1: 4}}
            }
        }, topology['objects'])

    def test_dedup_when_a_line_abdcd_self_intersects_with_its_end_it_is_cut_into_abd_dcd(self):
        topology = self.dedup(self.cut(self.extract({
            'abdcd': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [4, 0], [3, 0], [4, 0]]
            }
        })))

        self.assertDictEqual({
            'abdcd': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 2, 'next': {0: 2, 1: 4}}
            }
        }, topology['objects'])

    def test_dedup_when_an_old_line_abcdbe_self_intersects_and_shares_a_point_b_abcdbe_is_cut_into_ab_bcdb_be_and_fbg_is_cut_into_fb_bg(
            self):
        topology = self.dedup(self.cut(self.extract({
            'abcdbe': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [2, 0], [3, 0], [1, 0], [4, 0]]
            },
            'fbg': {
                'type': 'LineString',
                'arcs': [[0, 1], [1, 0], [2, 1]]
            }
        })))

        self.assertDictEqual({
            'abcdbe': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 1, 'next': {0: 1, 1: 4, 'next': {0: 4, 1: 5}}}
            },
            'fbg': {
                'type': 'LineString',
                'arcs': {0: 6, 1: 7, 'next': {0: 7, 1: 8}}
            }
        }, topology['objects'])

    def test_dedup_when_a_line_abca_is_closed_there_are_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'abca': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [0, 1], [0, 0]]
            }
        })))

        self.assertDictEqual({
            'abca': {
                'type': 'LineString',
                'arcs': {0: 0, 1: 3}
            }
        }, topology['objects'])

    def test_dedup_when_a_ring_abca_is_closed_there_are_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'abca': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [0, 1], [0, 0]]]
            }
        })))

        self.assertDictEqual({
            'abca': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            }
        }, topology['objects'])

    def test_dedup_exact_duplicate_rings_abca_and_abca_have_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'abca': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [0, 1], [0, 0]]]
            },
            'abca2': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [0, 1], [0, 0]]]
            }
        })))

        self.assertDictEqual({
            'abca': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            },
            'abca2': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            }
        }, topology['objects'])

    def test_dedup_reversed_duplicate_rings_abca_and_acba_have_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'abca': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [0, 1], [0, 0]]]
            },
            'acba': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [0, 1], [1, 0], [0, 0]]]
            }
        })))

        self.assertDictEqual({
            'abca': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            },
            'acba': {
                'type': 'Polygon',
                'arcs': [{0: 3, 1: 0}]
            }
        }, topology['objects'])

    def test_dedup_coincident_rings_abca_and_bcab_have_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'abca': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [0, 1], [0, 0]]]
            },
            'bcab': {
                'type': 'Polygon',
                'arcs': [[[1, 0], [0, 1], [0, 0], [1, 0]]]
            }
        })))

        self.assertDictEqual({
            'abca': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            },
            'bcab': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            }
        }, topology['objects'])

    def test_dedup_coincident_reversed_rings_abca_and_bacb_have_no_cuts(self):
        topology = self.dedup(self.cut(self.extract({
            'abca': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [0, 1], [0, 0]]]
            },
            'bacb': {
                'type': 'Polygon',
                'arcs': [[[1, 0], [0, 0], [0, 1], [1, 0]]]
            }
        })))

        self.assertDictEqual({
            'abca': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            },
            'bacb': {
                'type': 'Polygon',
                'arcs': [{0: 3, 1: 0}]
            }
        }, topology['objects'])

    def test_dedup_coincident_rings_abcda_efae_and_ghcg_are_cut_into_abc_cda_efae_and_ghcg(self):
        topology = self.dedup(self.cut(self.extract({
            'abcda': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]
            },
            'efae': {
                'type': 'Polygon',
                'arcs': [[[0, -1], [1, -1], [0, 0], [0, -1]]]
            },
            'ghcg': {
                'type': 'Polygon',
                'arcs': [[[0, 2], [1, 2], [1, 1], [0, 2]]]
            }
        })))

        self.assertDictEqual({
            'abcda': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 2, 'next': {0: 2, 1: 4}}]
            },
            'efae': {
                'type': 'Polygon',
                'arcs': [{0: 5, 1: 8}]
            },
            'ghcg': {
                'type': 'Polygon',
                'arcs': [{0: 9, 1: 12}]
            }
        }, topology['objects'])

    def test_dedup_coincident_rings_abca_and_dbed_have_no_cuts_but_are_rotated_to_share_b(self):
        topology = self.dedup(self.cut(self.extract({
            'abca': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [0, 1], [0, 0]]]
            },
            'dbed': {
                'type': 'Polygon',
                'arcs': [[[2, 1], [1, 0], [2, 2], [2, 1]]]
            }
        })))

        self.assertDictEqual({
            'abca': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 3}]
            },
            'dbed': {
                'type': 'Polygon',
                'arcs':  [{0: 4, 1: 7}]
            }
        }, topology['objects'])
        self.assertListEqual(topology['coordinates'][:4], [[1, 0], [0, 1], [0, 0], [1, 0]])
        self.assertListEqual(topology['coordinates'][4:], [[1, 0], [2, 2], [2, 1], [1, 0]])

    def test_dedup_overlapping_rings_abcda_and_befcb_are_cut_into_bc_cdab_and_befc_cb(self):
        topology = self.dedup(self.cut(self.extract({
            'abcda': {
                'type': 'Polygon',
                'arcs': [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]  # rotated to BCDAB, cut BC-CDAB
            },
            'befcb': {
                'type': 'Polygon',
                'arcs': [[[1, 0], [2, 0], [2, 1], [1, 1], [1, 0]]]
            }
        })))

        self.assertDictEqual({
            'abcda': {
                'type': 'Polygon',
                'arcs': [{0: 0, 1: 1, 'next': {0: 1, 1: 4}}]
            },
            'befcb': {
                'type': 'Polygon',
                'arcs': [{0: 5, 1: 8, 'next': {0: 1, 1: 0}}]
            }
        }, topology['objects'])



# -*- coding: utf-8 -*-

# Standard library imports
import json
import math
import unittest

# Third-party imports

# Application-specific imports
from pytopojson import (
    feature,
    mesh,
    stitch
)


class MeshTestCase(unittest.TestCase):
    def setUp(self):
        self.feature = feature.Feature()
        self.mesh = mesh.Mesh()
        self.stitch = stitch.Stitch()

    def test_mesh_ignores_null_geometries(self):
        topology = {
            "type": "Topology",
            "objects": {},
            "arcs": []
        }

        self.assertDictEqual(self.mesh(topology, [{"type": None}]), {
            "type": "MultiLineString",
            "coordinates": []
        })

    def test_mesh_stitches_together_two_connected_line_strings(self):
        topology = {
            "type": "Topology",
            "objects": {
                "collection": {
                    "type": "GeometryCollection",
                    "geometries": [
                        {"type": "LineString", "arcs": [0]},
                        {"type": "LineString", "arcs": [1]}
                    ]
                }
            }, "arcs": [
                [[1, 0], [2, 0]],
                [[0, 0], [1, 0]]
            ]
        }
        self.assertDictEqual({
            "type": "MultiLineString",
            "coordinates": [[[0, 0], [1, 0], [2, 0]]]
        }, self.mesh(topology, topology['objects']['collection']))

    def test_mesh_does_not_stitch_together_two_disconnected_line_strings(self):
        topology = {
            "type": "Topology",
            "objects": {
                "collection": {
                    "type": "GeometryCollection",
                    "geometries": [
                        {"type": "LineString", "arcs": [0]},
                        {"type": "LineString", "arcs": [1]}
                    ]
                }
            }, "arcs": [
                [[2, 0], [3, 0]],
                [[0, 0], [1, 0]]
            ]
        }
        self.assertDictEqual({
            "type": "MultiLineString",
            "coordinates": [[[2, 0], [3, 0]], [[0, 0], [1, 0]]]
        }, self.mesh(topology, topology['objects']['collection']))




class Geometry(object):
    def __init__(self):
        pass

    def __call__(self, inputs, *args, **kwargs):
        self.output = dict()
        for k, v in inputs.items():
            self.output[k] = self.geomify_object(inputs[k])

        return self.output

    def geomify_object(self, input):
        if input is None:
            return {'type': None}
        elif input['type'] == 'FeatureCollection':
            return self.geomify_feature_collection(input)
        elif input['type'] == 'Feature':
            return self.geomify_feature(input)
        else:
            return self.geomify_geometry(input)

    def geomify_feature_collection(self, input):
        output = dict()
        output['type'] = 'GeometryCollection'
        output['geometries'] = list(map(self.geomify_feature, input['features']))
        if input.get('bbox', None) is not None:
            output['bbox'] = input['bbox']

        return output

    def geomify_feature(self, input):
        output = self.geomify_geometry(input['geometry'])
        if input.get('id', None):
            output['id'] = input['id']
        if input.get('bbox', None) is not None:
            output['bbox'] = input['bbox']
        if input.get('properties', None):
            output['properties'] = input['properties']

        return output

    def geomify_geometry(self, input):
        if input is None:
            return {'type': None}

        output = dict()
        if input['type'] == 'GeometryCollection':
            output['type'] = 'GeometryCollection'
            output['geometries'] = list(map(self.geomify_geometry, input['geometries']))
        elif input['type'] == 'Point' or input['type'] == 'MultiPoint':
            output['type'] = input['type']
            output['coordinates'] = input['coordinates']
        else:
            output['type'] = input['type']
            output['arcs'] = input['coordinates']

        if input.get('bbox', None) is not None:
            output['bbox'] = input['bbox']

        return output



# -*- coding: <encoding name> -*-

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytopojson",
    version="1.0.0",
    author="Facundo Ferrín",
    author_email="facundo.ferrin@gmail.com",
    description="An extension to GeoJSON that encodes topology",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fferrin/pytopojson",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
      'numpy >= 1.15.0',
    ],
    python_requires='>=3.7.3',
)


# -*- coding: utf-8 -*-

# Standard library imports

# Third-party imports

# Application-specific imports
from pytopojson import (
    commons,
    feature,
    stitch
)


class ExtractArcs(object):
    def __init__(self):
        self.arcs = list()
        self.geoms_by_arc = list()
        self.geom = None

    def __call__(self, topology, object, filter, *args, **kwargs):
        self.geometry(object)

        for geoms in self.geoms_by_arc:
            if filter is None:
                self.arcs.append(geoms[0]['i'])
            else:
                if filter(geoms[0]['g'], geoms[-1]['g']):
                    self.arcs.append(geoms[0]['i'])

        return self.arcs

    def extract_0(self, i):
        j = ~i if i < 0 else i
        self.geoms_by_arc.get(j, []).append({
            'i': i,
            'g': self.geom
        })

    def extract_1(self, arcs):
        for arc in arcs:
            self.extract_0(arc)

    def extract_2(self, arcs):
        for arc in arcs:
            self.extract_1(arc)

    def extract_3(self, arcs):
        for arc in arcs:
            self.extract_2(arc)

    def geometry(self, o):
        self.geom = o
        _type = o['type']
        if _type == 'GeometryCollection':
            for g in o['geometries']:
                self.geometry(g)
        elif _type == 'LineString':
            self.extract_1(o['arcs'])
        elif _type == 'MultiLineString' or _type == 'Polygon':
            self.extract_2(o['arcs'])
        elif _type == 'MultiPolygon':
            self.extract_3(o['arcs'])


class MeshArcs(object):
    def __init__(self):
        self.stitch = stitch.Stitch()
        self.extract_arcs = ExtractArcs()

    def __call__(self, topology, obj=None, filt=None, *args, **kwargs):
        if obj is not None and filt is not None:
            arcs = self.extract_arcs(topology, obj, filt)
        else:
            n = len(topology['arcs'])
            arcs = [i for i in range(n)]
        return {
            'type': 'MultiLineString',
            'arcs': self.stitch(topology, arcs)
        }


class Mesh(object):
    def __init__(self):
        self.object = feature.Object()
        self.mesh_arcs = MeshArcs()

    def __call__(self, topology, *args, **kwargs):
        arcs = self.mesh_arcs(topology, *args)
        return self.object(topology, arcs)



# -*- coding: utf-8 -*-

# Standard library imports
import unittest

# Third-party imports

# Application-specific imports
from pytopojson import bounds


class BoundsTestCase(unittest.TestCase):
    def setUp(self):
        self.bbox = bounds.BoundingBox()

    def test_bounds_computes_bounding_box(self):
        foo = {
            'bar': {
                'type': 'LineString',
                'arcs': [[0, 0], [1, 0], [0, 2], [0, 0]]
            }
        }

        self.assertCountEqual(self.bbox(foo), [0, 0, 1, 2])

    def test_bounds_considers_points_as_well_as_arcs(self):
        foo = {
            'bar': {
                'type': 'MultiPoint',
                'coordinates': [[0, 0], [1, 0], [0, 2], [0, 0]]
            }
        }

        self.assertCountEqual(self.bbox(foo), [0, 0, 1, 2])


# -*- coding: utf-8 -*-

# Standard library imports
import json
import math
import unittest

# Third-party imports

# Application-specific imports
from pytopojson import feature


class FeatureTestCase(unittest.TestCase):
    def setUp(self):
        self.feature = feature.Feature()

    @staticmethod
    def _load_json_file(filepath):
        with open(filepath, 'r') as src:
            return json.load(src)

    @staticmethod
    def simple_topology(obj):
        return {
            'type': 'Topology',
            'transform': {
                'scale': [1, 1],
                'translate': [0, 0]
            },
            'objects': {
                'foo': obj
            },
            'arcs': [
                [[0, 0], [1, 0], [0, 1], [-1, 0], [0, -1]],
                [[0, 0], [1, 0], [0, 1]],
                [[1, 1], [-1, 0], [0, -1]],
                [[1, 1]],
                [[0, 0]]
            ]
        }

    def test_feature_the_geometry_type_is_preserved(self):
        t = self.simple_topology({'type': 'Polygon', 'arcs': [[0]]})
        self.assertEqual(
            self.feature(t, t['objects']['foo'])['geometry']['type'],
            'Polygon'
        )

    def test_feature_point_is_a_valid_geometry_type(self):
        t = self.simple_topology({'type': 'Point', 'coordinates': [0, 0]})
        self.assertDictEqual({
            'type': 'Feature',
            'properties': {},
            'geometry': {
                'type': 'Point',
                'coordinates': [0, 0]
            }
        }, self.feature(t, t['objects']['foo']))

    def test_feature_MultiPoint_is_a_valid_geometry_type(self):
        t = self.simple_topology({
            "type": "MultiPoint",
            "coordinates": [[0, 0]]
        })
        self.assertDictEqual({
            "type": "Feature",
            "properties": {},
            "geometry": {
                "type": "MultiPoint",
                "coordinates": [[0, 0]]
            }
        }, self.feature(t, t['objects']['foo']))

    def test_feature_LineString_is_a_valid_geometry_type(self):
        t = self.simple_topology({"type": "LineString", "arcs": [0]})
        self.assertDictEqual({
            "type": "Feature",
            "properties": {},
            "geometry": {
                "type": "LineString",
                "coordinates": [[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]
            }
        }, self.feature(t, t['objects']['foo']))

    def test_feature_MultiLineString_is_a_valid_geometry_type(self):
        t = self.simple_topology({"type": "MultiLineString", "arcs": [[0]]})
        self.assertDictEqual({
            "type": "Feature",
            "properties": {},
            "geometry": {
                "type": "MultiLineString",
                "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]
            }
        }, self.feature(t, t['objects']['foo']))

    def test_feature_line_strings_have_at_least_two_coordinates(self):
        t = self.simple_topology({"type": "LineString", "arcs": [3]})
        self.assertDictEqual({
            "type": "Feature",
            "properties": {},
            "geometry": {
                "type": "LineString",
                "coordinates": [[1, 1], [1, 1]]
            }
        }, self.feature(t, t['objects']['foo']))

        t = self.simple_topology({"type": "MultiLineString", "arcs": [[3], [4]]})
        self.assertDictEqual({
            "type": "Feature",
            "properties": {},
            "geometry": {
                "type": "MultiLineString",
                "coordinates": [[[1, 1], [1, 1]], [[0, 0], [0, 0]]]
            }
        }, self.feature(t, t['objects']['foo']))

    def test_feature_Polygon_is_a_valid_feature_type(self):
        t = self.simple_topology({"type": "Polygon", "arcs": [[0]]})
        self.assertDictEqual({
            "type": "Feature",
            "properties": {},
            "geometry": {
                "type": "Polygon",
                "coordinates": [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]
            }
        }, self.feature(t, t['objects']['foo']))

    def test_feature_MultiPolygon_is_a_valid_feature_type(self):
        t = self.simple_topology({"type": "MultiPolygon", "arcs": [[[0]]]})
        self.assertDictEqual({
            "type": "Feature",
            "properties": {},
            "geometry": {
                "type": "MultiPolygon",
                "coordinates": [[[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]]
            }
        }, self.feature(t, t['objects']['foo']))

    def test_feature_polygons_are_closed_with_at_least_four_coordinates(self):
        topology = {
            "type": "Topology",
            "transform": {
                "scale": [1, 1],
                "translate": [0, 0]
            }, "objects": {
                "foo": {
                    "type": "Polygon",
                    "arcs": [[0]]
                }, "bar": {
                    "type": "Polygon",
                    "arcs": [[0, 1]]
                }
            },
            "arcs": [[[0, 0], [1, 1]], [[1, 1], [-1, -1]]]
        }
        self.assertListEqual(
            self.feature(topology, topology['objects']['foo'])['geometry']['coordinates'],
            [[[0, 0], [1, 1], [0, 0], [0, 0]]]
        )
        self.assertListEqual(
            self.feature(topology, topology['objects']['bar'])['geometry']['coordinates'],
            [[[0, 0], [1, 1], [0, 0], [0, 0]]]
        )

    def test_feature_top_level_geometry_collections_are_mapped_to_feature_collections(self):
        t = self.simple_topology({
            "type": "GeometryCollection",
            "geometries": [{
                "type": "MultiPolygon",
                "arcs": [[[0]]]
            }]
        })
        self.assertDictEqual({
            "type": "FeatureCollection",
            "features": [{
                "type": "Feature",
                "properties": {},
                "geometry": {
                    "type": "MultiPolygon",
                    "coordinates": [[[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]]
                }
            }]
        }, self.feature(t, t['objects']['foo']))

    def test_feature_geometry_collections_can_be_nested(self):
        t = self.simple_topology({
            "type": "GeometryCollection",
            "geometries": [{
                "type": "GeometryCollection",
                "geometries": [{
                    "type": "Point",
                    "coordinates": [0, 0]
                }]
            }]
        })
        self.assertDictEqual({
            "type": "FeatureCollection",
            "features": [{
                "type": "Feature",
                "properties": {},
                "geometry": {
                    "type": "GeometryCollection",
                    "geometries": [{
                        "type": "Point",
                        "coordinates": [0, 0]
                    }]
                }
            }]
        }, self.feature(t, t['objects']['foo']))
    
    def test_feature_top_level_geometry_collections_do_not_have_ids_but_second_level_geometry_collections_can(self):
        t = self.simple_topology({
            "type": "GeometryCollection",
            "id": "collection",
            "geometries": [{
                "type": "GeometryCollection",
                "id": "feature",
                "geometries": [{
                    "type": "Point",
                    "id": "geometry",
                    "coordinates": [0, 0]
                }]
            }]
        })
        self.assertDictEqual({
            "type": "FeatureCollection",
            "features": [{
                "type": "Feature",
                "id": "feature",
                "properties": {},
                "geometry": {
                    "type": "GeometryCollection",
                    "geometries": [{
                        "type": "Point",
                        "coordinates": [0, 0]
                    }]
                }
            }]
        }, self.feature(t, t['objects']['foo']))
    
    def test_feature_top_level_geometry_collections_do_not_have_properties_but_second_level_geometry_collections_can(self):
        t = self.simple_topology({
            "type": "GeometryCollection",
            "properties": {
                "collection": True
            }, "geometries": [{
                "type": "GeometryCollection",
                "properties": {
                    "feature": True
                },
                "geometries": [{
                    "type": "Point",
                    "properties": {
                        "geometry": True
                    }, "coordinates": [0, 0]
                }]
            }]
        })
        self.assertDictEqual({
            "type": "FeatureCollection",
            "features": [{
                "type": "Feature",
                "properties": {
                    "feature": True
                }, "geometry": {
                    "type": "GeometryCollection",
                    "geometries": [{
                        "type": "Point",
                        "coordinates": [0, 0]
                    }]
                }
            }]
        }, self.feature(t, t['objects']['foo']))
    
    def test_feature_the_object_id_is_promoted_to_feature_id(self):
        t = self.simple_topology({
            "id": "foo",
            "type": "Polygon",
            "arcs": [[0]]
        })
        self.assertEqual("foo", self.feature(t, t['objects']['foo'])['id'])

    def test_feature_any_object_properties_are_promoted_to_feature_properties(self):
        t = self.simple_topology({
            "type": "Polygon", 
            "properties": {
                "color": "orange", 
                "size": 42
            }, "arcs": [[0]]
        })
        self.assertDictEqual({
            "color": "orange", 
            "size": 42
        }, self.feature(t, t['objects']['foo'])['properties'])

    def test_feature_the_object_id_is_optional(self):
        t = self.simple_topology({
            "type": "Polygon", 
            "arcs": [[0]]
        })
        self.assertEqual(
            self.feature(t, t['objects']['foo']).get('id', None), 
            None
        )

    def test_feature_object_properties_are_created_if_missing(self):
        t = self.simple_topology({
            "type": "Polygon", 
            "arcs": [[0]]
        })
        self.assertDictEqual(
            self.feature(t, t['objects']['foo'])['properties'], 
            {}
        )

    def test_feature_arcs_are_converted_to_coordinates(self):
        t = self.simple_topology({
            "type": "Polygon",
            "arcs": [[0]]
        })
        self.assertListEqual(
            self.feature(t, t['objects']['foo'])['geometry']['coordinates'],
            [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]]
        )

    def test_feature_negative_arc_indexes_indicate_reversed_coordinates(self):
        t = self.simple_topology({
            "type": "Polygon",
            "arcs": [[~0]]
        })
        self.assertListEqual(
            self.feature(t, t['objects']['foo'])['geometry']['coordinates'],
            [[[0, 0], [0, 1], [1, 1], [1, 0], [0, 0]]]
        )
    
    def test_feature_when_multiple_arc_indexes_are_specified_coordinates_are_stitched_together(self):
        t = self.simple_topology({
            "type": "LineString",
            "arcs": [1, 2]
        })
        self.assertListEqual(
            self.feature(t, t['objects']['foo'])['geometry']['coordinates'],
            [[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]
        )

        t = self.simple_topology({
            "type": "Polygon",
            "arcs": [[~2, ~1]]
        })
        self.assertListEqual(
            self.feature(t, t['objects']['foo'])['geometry']['coordinates'],
            [[[0, 0], [0, 1], [1, 1], [1, 0], [0, 0]]]
        )

    def test_feature_unknown_geometry_types_are_converted_to_null_geometries(self):
        topology = {
            "type": "Topology",
            "transform": {
                "scale": [1, 1],
                "translate": [0, 0]
            }, "objects": {
                "foo": {"id": "foo"},
                "bar": {
                    "type": "Invalid",
                    "properties": {"bar": 2}
                },
                "baz": {
                    "type": "GeometryCollection",
                    "geometries": [{
                        "type": "Unknown",
                        "id": "unknown"
                    }]
                }
            },
            "arcs": []
        }
        self.assertDictEqual({
            "type": "Feature",
            "id": "foo",
            "properties": {},
            "geometry": None
        }, self.feature(topology, topology['objects']['foo']))
        self.assertDictEqual({
            "type": "Feature",
            "properties": {"bar": 2},
            "geometry": None
        }, self.feature(topology, topology['objects']['bar']))
        self.assertDictEqual({
            "type": "FeatureCollection",
            "features": [{
                "type": "Feature",
                "id": "unknown",
                "properties": {},
                "geometry": None
            }]
        }, self.feature(topology, topology['objects']['baz']))

    def test_feature_preserves_additional_dimensions_in_Point_geometries(self):
        t = {
            "type": "Topology",
            "objects": {
                "point": {
                    "type": "Point",
                    "coordinates": [1, 2, "foo"]
                }
            }, "arcs": []
        }
        self.assertDictEqual({
            "type": "Feature",
            "properties": {},
            "geometry": {
                "type": "Point",
                "coordinates": [1, 2, "foo"]
            }
        }, self.feature(t, t['objects']['point']))

    def test_feature_preserves_additional_dimensions_in_MultiPoint_geometries(self):
        t = {
            "type": "Topology",
            "objects": {
                "points": {
                    "type": "MultiPoint",
                    "coordinates": [[1, 2, "foo"]]
                }
            }, "arcs": []
        }
        self.assertDictEqual({
            "type": "Feature",
            "properties": {},
            "geometry": {
                "type": "MultiPoint",
                "coordinates": [[1, 2, "foo"]]
            }
        }, self.feature(t, t['objects']['points']))

    def test_feature_preserves_additional_dimensions_in_LineString_geometries(self):
        t = {
            "type": "Topology",
            "objects": {
                "line": {
                    "type": "LineString",
                    "arcs": [0]
                }
            }, "arcs": [[[1, 2, "foo"], [3, 4, "bar"]]]
        }
        self.assertDictEqual({
            "type": "Feature",
            "properties": {},
            "geometry": {
                "type": "LineString",
                "coordinates": [[1, 2, "foo"], [3, 4, "bar"]]
            }
        }, self.feature(t, t['objects']['line']))



# -*- coding: utf-8 -*-

# Standard library imports
import math
from collections.abc import Iterable

# Third-party imports

# Application-specific imports
from pytopojson import (
    bbox,
    commons,
    untransform,
)


class Quantize(object):
    def __init__(self):
        self.bbox = bbox.BBox()
        self.untransform = untransform.Untransform()
        self.t = None

    def __call__(self, topology, transform=None, *args, **kwargs):
        if topology.get('transform', None) is not None:
            raise ValueError('Already quantized.')

        no_transform = transform is None
        no_scale = 'scale' not in transform if isinstance(transform, Iterable) else True
        if no_transform or no_scale:
            if transform is None or math.floor(transform) < 2:
                raise ValueError('n must be ≥2.')
            n = math.floor(transform)

            if 'bbox' in topology and topology['bbox'] is not None:
                box = topology['bbox']
            else:
                box = self.bbox(topology)

            x_0, y_0, x_1, y_1 = box
            transform = {
                'scale': [
                    (x_1 - x_0) / (n - 1) if x_0 < x_1 else 1,
                    (y_1 - y_0) / (n - 1) if y_0 < y_1 else 1
                ],
                'translate': [x_0, y_0]
            }
        else:
            box = topology['bbox']

        self.t = self.untransform(transform)
        inputs = topology['objects']
        outputs = dict()

        for k in inputs:
            outputs[k] = self.quantize_geometry(inputs[k])

        return {
            'type': 'Topology',
            'bbox': box,
            'transform': transform,
            'objects': outputs,
            'arcs': list(map(lambda x: self.quantize_arc(x), topology['arcs']))
        }

    def quantize_point(self, point):
        return self.t(point)

    def quantize_geometry(self, input):
        if input['type'] == 'GeometryCollection':
            output = {
                'type': 'GeometryCollection',
                'geometries': list(map(lambda x: self.quantize_geometry(x),
                                       input['geometries']))
            }
        elif input['type'] == 'Point':
            output = {
                'type': 'Point',
                'coordinates': self.quantize_point(input['coordinates'])
            }
        elif input['type'] == 'MultiPoint':
            output = {
                'type': 'MultiPoint',
                'coordinates': list(map(lambda x: self.quantize_point(x),
                                       input['coordinates']))
            }
        else:
            return input

        if 'id' in input:
            output['id'] = input['id']
        if 'bbox' in input:
            output['bbox'] = input['bbox']
        if 'properties' in input:
            output['properties'] = input['properties']

        return output

    def quantize_arc(self, input):
        i = 1
        j = 1
        n = len(input)
        output = commons.Array(n)   # Pessimistic
        output[0] = self.t(input[0], 0)

        while i < n:
            p = self.t(input[i], i)
            if p[0] or p[1]:
                output[j] = p    # Non-coincident points
                j += 1
            i += 1

        if j == 1:
            output[j] = [0, 0]   # An arc must have at least two points
            j += 1

        output = output[:j]
        return output
